import greenfoot.*;  // (World, Actor, GreenfootImage, Greenfoot and MouseInfo)
import java.util.*;  // (ArrayList and List)

/**
 * A ScrollingWorld maintains another world inside of it and displays only a small portion of that
 * world at a time. The unfortunate tradeoff for this power is that actors in the underlying world
 * cannot register mouse clicks. It will also ignore the underlying world's background and substitute
 * its own, since the underlying world will usually be very large, and it will be more memory-
 * effective -- and for that matter, more memory-possible -- without much tradeoff for visual
 * effect -- to display a fixed background the size of the window, instead of a scrolling background
 * the size of the entire underlying world.
 * 
 * HOW TO USE A SCROLLING WORLD:
 * - On construction, add actors into getUnderlying() instead of directly into the world.
 * -- Unless your actor is part of the user interface and you want it to always show up
 *    in the same place in the window and be clicked on by the player and stuff. Then you can
 *    add it directly into the world.
 * - If you're modifying the background, use setPersBackground() and getPersBackground() instead of
 *   modifying the background proper, because the background proper gets reset to the persistent
 *   background every frame before being populated with the actor images, so as to avoid the infamous
 *   hall of mirrors effect.
 * - If you add an actor into getUnderlying(), that will be its getWorld(). If you want to get the
 *   actual ScrollingWorld that's currently being used by Greenfoot as the current world, you'll
 *   have to do ((ScrollingWorld.ScrolledWorld) getWorld()).getOverlaying(). The dot in the name of
 *   the class you're casting to is because ScrolledWorld is a class defined *inside* of ScrollingWorld.
 * 
 * @author BenMcBen
 * @version 10/26/2014
 */
public abstract class ScrollingWorld extends World {
    /**
     * A ScrolledWorld is really just a basic world, except that it isn't abstract and a reference
     * to its corresponding ScrollingWorld can be obtained from it. It also allows access to some
     * data representing its act and paint orders, even though the act and paint orders themselves
     * cannot be accessed directly.
     */
    public class ScrolledWorld extends World {
        private ArrayList<Class> actOrder;
        private ArrayList<Class> paintOrder;
        /**
         * constructor for ScrolledWorld: construct a world and the external act and paint orders
         */
        public ScrolledWorld(int width, int height, int cell) {
            super(width, height, cell);
            actOrder = new ArrayList<Class>();
            paintOrder = new ArrayList<Class>();
        }
        /**
         * getOverlaying - return the ScrollingWorld that's scrolling this ScrolledWorld
         */
        public ScrollingWorld getOverlaying() {
            return ScrollingWorld.this;
        }
        /**
         * getActOrder - return the list of classes last supplied to setActOrder
         */
        public List<Class> getActOrder() {
            return actOrder;
        }
        /**
         * getPaintOrder - return the list of classes last supplied to setPaintOrder
         */
        public List<Class> getPaintOrder() {
            return paintOrder;
        }
        /**
         * setActOrder - set the act order normally, and also set the more superficial act order
         * stored in this class
         */
        public void setActOrder(Class ... classes) {
            super.setActOrder(classes);
            actOrder = new ArrayList<Class>(Arrays.asList(classes));
        }
        /**
         * setPaintOrder - set the paint order normally, and also set the more superficial paint
         * order stored in this class
         */
        public void setPaintOrder(Class ... classes) {
            super.setPaintOrder(classes);
            paintOrder = new ArrayList<Class>(Arrays.asList(classes));
        }
    }
    private int scrollx;
    private int scrolly;
    private ScrolledWorld underlying;
    private GreenfootImage persistentBackground;
    /**
     * constructor for ScrollingWorld - create underlying world and scroll area of specified
     * dimensions
     */
    public ScrollingWorld(int width, int height, int swidth, int sheight) {
        super(swidth, sheight, 1);
        underlying = new ScrolledWorld(width, height, 1);
        scrollx = 0;
        scrolly = 0;
        makePersistentBackground();
        updateBackground();
    }
    /**
     * smaller constructor for ScrollingWorld - create scroll area of 600x400 for underlying
     * world of specified dimensions
     */
    public ScrollingWorld(int width, int height) {
        this(width, height, 600, 400);
    }
    /**
     * smallest constructor for ScrollingWorld - create scroll area of 600x400 for underlying
     * 5120x3840 world; this is 8x8 640x480 areas
     */
    public ScrollingWorld() {
        this(5120, 3840);
    }
    /**
     * getUnderlying - return the ScrolledWorld that this ScrollingWorld is scrolling
     */
    public ScrolledWorld getUnderlying() {
        return underlying;
    }
    /**
     * act - update the background and then defer to ScrolledWorld and its actors in the
     * act order
     */
    public void act() {
        updateBackground();
        // we need to keep track of what actors have already acted in case due to inheritance
        // the same actor comes up in the list twice or more
        ArrayList<Actor> alreadyActed = new ArrayList<Actor>();
        underlying.act();
        for (Class klass : underlying.getActOrder()) {
            for (Actor actor : (List<Actor>) underlying.getObjects(klass)) {
                if (!alreadyActed.contains(actor) && underlying.getObjects(klass).contains(actor)) {
                    actor.act();
                    alreadyActed.add(actor);
                }
            }
        }
        for (Actor actor : (List<Actor>) underlying.getObjects(null)) {
            if (!alreadyActed.contains(actor) && underlying.getObjects(null).contains(actor)) {
                actor.act();
                alreadyActed.add(actor);
            }
        }
    }
    /**
     * started - defer to ScrolledWorld
     */
    public void started() {
        underlying.started();
    }
    /**
     * stopped - defer to ScrolledWorld
     */
    public void stopped() {
        underlying.stopped();
    }
    /**
     * makePersistentBackground - make a background equivalent to the background right now
     * and store it for clean resets
     */
    private void makePersistentBackground() {
        persistentBackground = new GreenfootImage(getWidth(), getHeight());
        persistentBackground.drawImage(super.getBackground(), 0, 0);
    }
    /**
     * updateBackground - draw a new background reflecting what's going on in the underlying
     * world in the area bounded by the coordinates scrollx,scrolly and the dimensions
     * getWidth()xgetHeight(); ignore the underlying world's background and don't paint
     * anything that won't wind up on the screen anyway
     */
    private void updateBackground() {
        int screenx;
        int screeny;
        super.getBackground().clear();
        super.getBackground().drawImage(persistentBackground, 0, 0);
        // bg.drawImage(underlying.getBackground(), -scrollx, -scrolly);
        //// Ignoring the underlying world's background for memory conservation
        for (Actor actor : (List<Actor>) underlying.getObjects(null)) {
            tryPaintActor(actor);
        }
        // we need to reverse the paint order because according to the API the objects at
        // the beginning should be drawn on top AKA last
        List<Class> realPaintOrder = (List<Class>) ((ArrayList<Class>) underlying.getPaintOrder()).clone();
        Collections.reverse(realPaintOrder);
        for (Class klass : realPaintOrder) {
            for (Actor actor : (List<Actor>) underlying.getObjects(klass)) {
                tryPaintActor(actor);
            }
        }
    }
    /**
     * tryPaintActor - consider painting the actor, but give up if the actor would just wind
     * up offscreen anyway
     */
    private void tryPaintActor(Actor actor) {
        int max = Math.max(actor.getImage().getWidth(), actor.getImage().getHeight());
        GreenfootImage actorImage = new GreenfootImage(max, max);
        actorImage.drawImage(actor.getImage(), 0, 0);
        actorImage.rotate(actor.getRotation());
        int screenx = actor.getX() - actor.getImage().getWidth()/2 - scrollx;
        int screeny = actor.getY() - actor.getImage().getHeight()/2 - scrolly;
        if (screenx + actor.getImage().getWidth() < 0) return;
        if (screenx - actor.getImage().getWidth() > getWidth()) return;
        if (screeny + actor.getImage().getHeight() < 0) return;
        if (screeny - actor.getImage().getHeight() > getHeight()) return;
        super.getBackground().drawImage(actorImage, screenx, screeny);
    }
    /**
     * setPersBackground - set the persistent background
     */
    public void setPersBackground(GreenfootImage image) {
        persistentBackground = image;
    }
    /**
     * getPersBackground - get the persistent background
     */
    public GreenfootImage getPersBackground() {
        return persistentBackground;
    }
    /**
     * scroll - change the scrolling coordinates by a specified amount
     */
    public void scroll(int dx, int dy) {
        scrollTo(scrollx + dx, scrolly + dy);
    }
    /**
     * scrollTo - directly set the scrolling coordinates and ensure they're valid
     */
    public void scrollTo(int x, int y) {
        scrollx = x;
        scrolly = y;
        if (scrollx < 0) {
            scrollx = 0;
        }
        if (scrollx > underlying.getWidth() - getWidth()) {
            scrollx = underlying.getWidth() - getWidth();
        }
        if (scrolly < 0) {
            scrolly = 0;
        }
        if (scrolly > underlying.getHeight() - getHeight()) {
            scrolly = underlying.getHeight() - getHeight();
        }
        updateBackground();
    }
    /**
     * center - center the window around a position
     */
    public void center(int x, int y) {
        scrollTo(x - getWidth()/2, y - getHeight()/2);
    }
    /**
     * center - center the window around an actor
     */
    public void center(Actor actor) {
        center(actor.getX(), actor.getY());
    }
}
